/*
 * This software is for a dsPIC33f used as a multi-channel analyzer
 * for gamma spectroscopy.
 * Pulse amplitudes are read in and a spectrum is generated.
 * Commands can be sent, and data can be retrieved via the serial interface (RS-232 levels).
 */

#include <stdio.h>
#include <stdlib.h>
#include <p33Fxxxx.h>
#include <dsp.h>
#include <stdint.h>
#include "constants.h"
#include "config.h"
#include "adc.h"
#include "interrupts.h"
#include "LED.h"
#include "DBTimer.h"
#include "UART.h"
#include "Clock.h"

//configuration bits
_FOSCSEL(FNOSC_FRC);
_FOSC(FCKSM_CSECMD & OSCIOFNC_ON & POSCMD_NONE);
_FWDT(FWDTEN_OFF);
_FICD(JTAGEN_OFF & ICS_PGD3);

//function prototypes for ISRs:
void __attribute__((__interrupt__, __auto_psv__)) _INT0Interrupt(void);
void __attribute__((__interrupt__, __auto_psv__)) _ADC1Interrupt(void);
void __attribute__((__interrupt__, __auto_psv__)) _T1Interrupt(void);

void __attribute__((__interrupt__, __auto_psv__)) _U1RXInterrupt(void);
void __attribute__((__interrupt__, __auto_psv__)) _U1TXInterrupt(void);

void sendMsg(uint8_t type, uint8_t* msg);
void invokeBulk(uint16_t cnt);
uint8_t* getString(uint16_t data, uint8_t* out);
uint8_t* getStringLong(uint32_t data, uint8_t* out);
uint16_t getInt(uint8_t* str);

//global variables
__eds__ uint16_t __attribute__((section("eds1"),space(eds))) spec [SPECTRUM_LENGTH];
uint16_t lastCounts = 0;
uint16_t lastPulseHeight = 0;

uint8_t INTStatus = ARMED_FIRST;

uint8_t incomingByte = 0;
uint8_t inputBuffer[INPUT_BUFFER_LENGTH] = {0};
uint8_t outputBuffer[OUTPUT_BUFFER_LENGTH] = {0};
uint8_t dynamicMsgBuffer[10] = {0};
uint8_t OKMessage[] = {"OK"};
uint8_t ERRMessage[] = {"ERR"};
uint8_t* inputPos = inputBuffer;
uint8_t* outputPos = outputBuffer;

uint8_t bulkmode = OFF;
uint16_t bulk_pos = 0;

//general purpose
uint8_t done = 0;
uint8_t devnull = 0;        //trash
uint16_t i = 0;

uint32_t time = 0;

//ISRs:
//UART recieve interrupt
void __attribute__((__interrupt__, __auto_psv__)) _U1RXInterrupt(void)
{
    SET_LED_3;
    IFS0bits.U1RXIF = 0;

    //get all recieved bytes from fifo:
    while(U1STAbits.URXDA == 1)
    {
        incomingByte = U1RXREG;

        //check for end-of-line
        if(incomingByte != LF)
        {
            if((inputPos - inputBuffer < INPUT_BUFFER_LENGTH - 1) && (incomingByte != LF))
            {
                //store incoming character in buffer
                *inputPos = incomingByte;
                inputPos++;
            }
            else
            {
                //reset buffer
                inputPos = inputBuffer;
            }
        }
        else
        {
            //command recieved completely:
            *inputPos = 0;
            inputPos = inputBuffer;

            //parse command:
            switch(*inputPos)
            {
                case GET_CHANNEL:
                    if(getInt(inputPos + 1) < SPECTRUM_LENGTH)
                    {
                        getString(spec[getInt(inputPos + 1)], dynamicMsgBuffer);
                        sendMsg(OK, dynamicMsgBuffer);
                    }
                    else
                    {
                        sendMsg(ERROR, 0);
                    }
                    break;

                case GET_ALL_CHANNELS:
                    bulk_pos = 0;
                    invokeBulk(bulk_pos);
                    break;

                case RESET_ALL_CHANNELS:
                    for(i = 0; i < SPECTRUM_LENGTH; i++)
                    {
                        spec[i] = 0;
                    }
                    sendMsg(OK, 0);
                    break;

                case GET_LAST_CNTS:
                    getString(lastCounts, dynamicMsgBuffer);
                    sendMsg(OK, dynamicMsgBuffer);
                    break;

                case RESET_LAST_CNTS:
                    lastCounts = 0;
                    sendMsg(OK, 0);
                    break;

                case STOP_ACQ:
                    DISABLE_ACQ;   //disable external interrupt (INT0)
                    RESET_LINE = 0;
                    CLR_LED_2;
                    TICKOUT_LINE = 0;
                    sendMsg(OK, 0);
                    break;

                case START_ACQ:
                    ENABLE_ACQ;    //enable external interrupt (INT0)
                    sendMsg(OK, 0);
                    break;

                case GET_LAST_HEIGHT:
                    getString(lastPulseHeight, dynamicMsgBuffer);
                    sendMsg(OK, dynamicMsgBuffer);
                    break;

                case GET_TIME:
                    getStringLong(time, dynamicMsgBuffer);
                    sendMsg(OK, dynamicMsgBuffer);
                    break;

                default:
                    //error:
                    sendMsg(ERROR, 0);
                    break;
            }

            inputPos = inputBuffer;
        }
    }

    CLR_LED_3;
}

// UART transmit interrupt:
void __attribute__((__interrupt__, __auto_psv__)) _U1TXInterrupt(void)
{
    SET_LED_3;
    
    IFS0bits.U1TXIF = 0;

    //send output buffer:
    if(*outputPos != 0)
    {
        U1TXREG = *outputPos;
        outputPos++;
    }
    else
    {
        if(bulkmode == OFF)
        {
            outputPos = outputBuffer;   //reset ptr
        }
        else
        {
            outputPos = dynamicMsgBuffer;   //reset ptr
            if(bulk_pos < SPECTRUM_LENGTH - 1)
            {
                bulk_pos++;
                invokeBulk(bulk_pos);
            }
            else
            {
                bulkmode = OFF;
            }
        }
    }

    CLR_LED_3;
}

//T1 is for debouncing the INT signal
void __attribute__((__interrupt__, __auto_psv__)) _T1Interrupt(void)
{
    IFS0bits.T1IF = 0;

    //stop timer
    RESET_TIMER;
    STOP_TIMER;

    //T1 interrupt occured, debouncing finished
    if(INTStatus == ARMED)
    {
        //debounced positive edge - new pulse started
        //switch ADC S/H to "sampling"
        AD1CON1bits.SAMP = 1;

        //change INT0 polarity to "falling" edge
        INTCON2bits.INT0EP = 1;

        INTStatus = SAMPLE;
    }
    else if(INTStatus == WAIT_PULSE)
    {
        //debounced negative edge - pulse finished

        //change INT0 polarity back to "rising" edge
        INTCON2bits.INT0EP = 0;
        INTStatus = ARMED_FIRST;
    }
}

//AD-interrupt
void __attribute__((__interrupt__, __auto_psv__)) _ADC1Interrupt(void)
{
    IFS0bits.AD1IF = 0;

    //AD conversion finished, add new value to histogram
    spec[ADC1BUF0] += 1;
    lastPulseHeight = ADC1BUF0;
    lastCounts++;

    //reset holding capacitor:
    RESET_LINE = 1;

    //tickout signal
    TICKOUT_LINE = 1;

    CLR_LED_2;
}

//external interrupt (INT signal from peak-detector)
void __attribute__((__interrupt__, __auto_psv__)) _INT0Interrupt(void)
{
    SET_LED_2;

    IFS0bits.INT0IF = 0;    //reset interrupt flag

    //state machine
    if(INTStatus == ARMED_FIRST)
    {
        //this is the first positive edge, start debouncing timer:
        RESET_LINE = 0;
        TICKOUT_LINE = 0;
        START_TIMER;
        INTStatus = ARMED;
    }
    else if(INTStatus == ARMED)
    {
        //waiting for next pulse; debounce INT-Signal and swith internal ADC to "sampling"
        //reset Time when new positive edge occurs, when timer overflows, debouncing completed
        RESET_TIMER;
    }
    else if(INTStatus == SAMPLE)
    {
        //this is the first negative edge, sample peak value and wait for pulse to decay
        START_AD_CONVERSION;
        START_TIMER;
        INTStatus = WAIT_PULSE;
    }
    else if(INTStatus == WAIT_PULSE)
    {
        //negative edge, reset debouncing timer
        RESET_TIMER;
    }
}

int main()
{
    initOSC();  //40MIPS
    initADC();
    initLEDs();
    initUART();
    initDebouncingTimer();
    initClock();
    initInterrupts();

    TRISBbits.TRISB1 = 0;   //reset-pin
    RESET_LINE = 0;

    TRISBbits.TRISB12 = 0;  //tickout
    TICKOUT_LINE = 0;
    
    uint16_t i = 0;
    uint16_t j = 0;

    sendMsg(OK, 0);     //send OK-message after initialization

    while(1)
    {      
        //idle-loop LED blinking:
        setLED1();
        for(i = 0; i < 20000; i++)
        {
            for(j = 0; j < 50; j++)
            {
                
            }
        }

        clearLED1();
        for(i = 0; i < 50000; i++)
        {
            for(j = 0; j < 150; j++)
            {

            }
        }
    }

    while(1);
}

//auxiliary functions for parsing strings to integers and converting integers to strings
uint16_t getInt(uint8_t* str)
{
    uint8_t* start = str;
    uint16_t base_exp = 1;
    uint16_t outval = 0;
    
    while(*str != 0)
    {
        str++;
    }

    do
    {
        str--;
        outval += ((*str) - ASCII_OFFSET) * base_exp;
        base_exp *= 10;
    }
    while(start != str);

    return outval;
}

//convert a 16-bit integer to a string
uint8_t* getString(uint16_t data, uint8_t* out)
{
    uint16_t start_mag = 1;
    uint16_t tmp = 0;

    //get magnitude:
    while(data / start_mag >= 10)
    {
        start_mag *= 10;
    }

    while(start_mag > 0)
    {
        tmp = (uint16_t)(data / start_mag);
        *out = tmp + ASCII_OFFSET;
        data %= start_mag;
        start_mag /= 10;
        out++;
    }

    *out = 0;

    return out;
}

//convert a 32-bit unsigned integer to a string
uint8_t* getStringLong(uint32_t data, uint8_t* out)
{
    uint32_t start_mag = 1;
    uint32_t tmp = 0;

    //get magnitude:
    while(data / start_mag >= 10)
    {
        start_mag *= 10;
    }

    while(start_mag > 0)
    {
        tmp = (uint32_t)(data / start_mag);
        *out = tmp + ASCII_OFFSET;
        data %= start_mag;
        start_mag /= 10;
        out++;
    }

    *out = 0;

    return out;
}

//send next value when in bulk-transfer mode
void invokeBulk(uint16_t cnt)
{
    uint8_t* end_data;
    bulkmode = ON;
    end_data = getString(spec[cnt], dynamicMsgBuffer);

    *end_data = '\n';
    end_data++;
    *end_data = 0;

    outputPos = dynamicMsgBuffer;

    U1TXREG = *outputPos;
    outputPos++;
}

//sends message back to PC (comprising a static text ("OK" or "ERR") as well as a
//"dynamic" part (usually the value of some register/channel sent back)
void sendMsg(uint8_t type, uint8_t* msg)
{
    uint8_t* msgptr;
    if(type == OK)
        msgptr = OKMessage;
    else if(type == ERROR)
        msgptr = ERRMessage;

    outputPos = outputBuffer;

    //copy static message text into output buffer:
    while(*msgptr != 0)
    {
        *outputPos = *msgptr;
        outputPos++;
        msgptr++;
    }

    //add dynamic text:
    if(msg != 0)
    {    
        *outputPos = '\n';
        outputPos++;
        
        while(*msg != 0)
        {
            *outputPos = *msg;
            outputPos++;
            msg++;
        }
    }

    *outputPos = '\n';
    outputPos++;
    *outputPos = 0;

    //start sending
    outputPos = outputBuffer;
    U1TXREG = *outputPos;
    outputPos++;
}